use std::ffi::{c_char, CString};
use std::sync::OnceLock;

mod lua;
mod plugin;
mod stingray_sdk;

use plugin::Plugin;
use stingray_sdk::{GetApiFunction, LoggingApi, LuaApi, PluginApi, PluginApiID};

// TODO: Change these
/// The name that the plugin is registered to the engine as.
/// Must be unique across all plugins.
pub const PLUGIN_NAME: &str = "dt-plugin-template";
/// The module that Lua functions are assigned to.
pub const MODULE_NAME: &str = "DTPluginTemplate";

static PLUGIN: OnceLock<Plugin> = OnceLock::new();
static LOGGER: OnceLock<LoggingApi> = OnceLock::new();
static LUA: OnceLock<LuaApi> = OnceLock::new();

/// A macro to make accessing global statics a little more convenient.
#[macro_export]
macro_rules! global {
    ($name:ident) => {{
        if cfg!(debug_assertions) {
            $name.get().expect("global has not been initialized")
        } else {
            unsafe { $name.get().unwrap_unchecked() }
        }
    }};
}

#[no_mangle]
pub extern "C" fn get_name() -> *const c_char {
    let s = CString::new(PLUGIN_NAME).expect("Failed to create CString from plugin name");
    s.as_ptr()
}

#[no_mangle]
pub extern "C" fn setup_game(get_engine_api: GetApiFunction) {
    let logger = LOGGER.get_or_init(|| LoggingApi::get(get_engine_api));
    log::set_logger(logger).expect("Failed to set global logger.");
    log::set_max_level(log::LevelFilter::Info);

    let _ = LUA.get_or_init(|| LuaApi::get(get_engine_api));

    let plugin = PLUGIN.get_or_init(Plugin::new);
    plugin.setup_game();
}

#[no_mangle]
pub extern "C" fn shutdown_game() {
    let plugin = global!(PLUGIN);
    plugin.shutdown_game();
}

#[no_mangle]
pub extern "C" fn update_game(dt: f32) {
    let plugin = global!(PLUGIN);
    plugin.update_game(dt);
}

#[no_mangle]
pub extern "C" fn get_plugin_api(id: PluginApiID) -> *mut PluginApi {
    if id == PluginApiID::PLUGIN_API_ID {
        let api = PluginApi {
            get_name: Some(get_name),
            setup_game: Some(setup_game),
            update_game: Some(update_game),
            shutdown_game: Some(shutdown_game),
            ..Default::default()
        };

        Box::into_raw(Box::new(api))
    } else {
        std::ptr::null_mut()
    }
}
